
__author__ = 'Adam McKenna, https://github.com/amckenna41'
__authorEmail__ = 'amckenna41@qub.ac.uk'
__url__ = 'https://github.com/amckenna41/pySAR'
__credits__ = ['Adam McKenna']
__version__ = "2.0.4"
__license__ = 'MIT'
